(* :Title: IndexNotation.m *)
(* :Context: dhPark`IndexNotation` *)
(* :Author: Dal-Ho Park *)
(* :Summary: Manipulation of Indical Objects alike MathTensor *)
(* :Package Version: 1.1 *)
(* :Mathematica Version: 4.0 *)
(* :History:
    00.01.31: Includable library of IndexNotation
    00.02.04: absorbRuleWith for Absorb
    00.05.01: Re-structuring for publishing
    00.05.10: Off[Identity::argx] due to ExpandIndicalAtom or MMA Bug?
    00.05.18: Internal representation of TP-type: tpName[_List, _]
              for avoiding commutative pattern matching in Tensor package
    00.08.21: Removed ExpandIAtom in OrderingIndices
 *)
(* TODO
   Dnup is not justified in Spinor calculus! Use iFlagTable[DnupFlag]
 *)
(* Programming GUIDE
    Use transformation rules and functional programming instead of procedural programming.
    Use list-manipulating functons instead of loop functions that require indexing.
    Avoid the operations that change the size of lists, e.g., Append and Prepend.
    Avoid argument checking within a function by writing specific patterns.
    Attach rules to those symbols.
    Error and warning messages should be provided.
    New errors messages should be defined only if existing messages are inappropriate.
 *)
(* Note
      IndexNotationLib
         Driver: IndexNotation, IndexNotationPreDefine
         Applications:
            Tensor (including Kaluza-Klein, PPN Formalism)
            Spinor Calculus (including Super Gravity)?

      Index Type == ComponentType(Integer) | RegularType(Symbol) | DummyType(String) | ErrorType(Any Other)
         Direct input of Dummy indices can be dangerous. They should be generated only by the System!

      Dn/Up Flag: l* or u*

      Index Class
         Component
            Input
               Component_Integer : ...,-2,-1,1,2,...
            Output
               Component_Integer : 1,2,...

         Normal
            Input
               Regular_Symbol    : la,  ua,..., lo, uo
               Dummy_String      : l$1, u$1,...
            Output
               Regular           : a,..., o
               Dummy             : p,..., z, "$num",

         Greek
            Input
               Regular_Symbol    : lmu, umu,...
               Dummy_String      : l@1, u@2,...
            Output
               Regular           : mu,...
               Dummy             : zeta,..., "@num",

         Capital
            Input
               Regular_Symbol    : lA, uA,..., lO, uO
               Dummy_String      : l#1, u#2,...
            Output
               Regular           : A,..., O
               Dummy             : P,..., Z, "#num",

         Hat
            Input
               Regular_Symbol    : lha, uha,..., lho, uho
               Dummy_String      : l%1, u%2,...
            Output
               Regular           : OverHat[a],..., OverHat[o]
               Dummy             : OverHat[p],..., OverHat[z], "%num",

         Bar
            Input
               Regular_Symbol    : lba, uba,..., lbo, ubo
               Dummy_String      : l!1, u!2,...
            Output
               Regular           : OverBar[a],..., OverBar[a]
               Dummy             : OverBar[p],..., OverBar[z], "!num",

         Dot
            Input
               Regular_Symbol    : lda, uda,..., ldo, udo
               Dummy_String      : l~1, u~2,...
            Output
               Regular           : OverDot[a],..., OverDot[o]
               Dummy             : OverDot[p],..., OverDot[z], "~num",

         One : for an index in one-dimensional space
            Input
               Regular_Symbol    : l0, u0
               Dummy_String      : "l0", "u0"
            Output
               Regular           : 0
               Dummy             : "0"

      Indexed Object Class:
         iClassTable       = {{mainIndexClassQ,...}, {subIndexClassQ,...}}

      Indexed Object:
         iNameTable[oName] = {prtStr, nRank, permW, extra,...}
            permW = {{{1,2,..}, 1}, {{2,1,..}, -1},..}
            extra:
               {DefinedTensorQ}
               {DefinedOperatorQ, opType}
               {DefinedFormQ, nDegree}

               opType:
                  CD : opName[index, expr] like covariant derivative
                  LD : opName[objName, expr] like Lie derivative
                  XD : opName[expr] like exterior derivative
                  XP : opName[expr,...] like exterior product
                       This is printed by an user-supplied rutine
                  TP : opName[expr, expr] like tensor power

         oName[args]:
            objName[indices]
            opName[arg, expr,...]
            fNmae[indices]

      Absorb is very in-efficient when multi-g's are involved!
 *)

(*************** Unprotect Exported Symbols / Rutines ***************)
Unprotect[l0,la,lb,lc,ld,le,lf,lg,lh,li,lj,lk,ll,lm,ln,lo];
Unprotect[u0,ua,ub,uc,ud,ue,uf,ug,uh,ui,uj,uk,ul,um,un,uo];
Unprotect[lalpha,lbeta,lgamma,ldelta,lmu,lnu,lrho,lsigma,lepsilon,lkappa,llambda,ltau];
Unprotect[ualpha,ubeta,ugamma,udelta,umu,unu,urho,usigma,uepsilon,ukappa,ulambda,utau];
Unprotect[lA,lB,lC,lD,lE,lF,lG,lH,lI,lJ,lK,lL,lM,lN,lO];
Unprotect[uA,uB,uC,uD,uE,uF,uG,uH,uI,uJ,uK,uL,uM,uN,uO];
Unprotect[lha,lhb,lhc,lhd,lhe,lhf,lhg,lhh,lhi,lhj,lhk,lhl,lhm,lhn,lho];
Unprotect[uha,uhb,uhc,uhd,uhe,uhf,uhg,uhh,uhi,uhj,uhk,uhl,uhm,uhn,uho];
Unprotect[lba,lbb,lbc,lbd,lbe,lbf,lbg,lbh,lbi,lbj,lbk,lbl,lbm,lbn,lbo];
Unprotect[uba,ubb,ubc,ubd,ube,ubf,ubg,ubh,ubi,ubj,ubk,ubl,ubm,ubn,ubo];
Unprotect[lda,ldb,ldc,ldd,lde,ldf,ldg,ldh,ldi,ldj,ldk,ldl,ldm,ldn,ldo];
Unprotect[uda,udb,udc,udd,ude,udf,udg,udh,udi,udj,udk,udl,udm,udn,udo];
(********************************************************************)

BeginPackage["dhPark`IndexNotation`"]

Needs["dhPark`dhUtilities`"]

(***** Usage messages for the exported functions *****)
dbg::usage = ""

(* Indices *)
BarIndexQ::usage       = "BarIndexQ[aIndex]"
CapitalIndexQ::usage   = "CapitalIndexQ[aIndex]"
ComponentIndexQ::usage = "ComponentIndexQ[aIndex]"
DnIndexQ::usage        = "DnIndexQ[aIndex, <testQ>]"
DotIndexQ::usage       = "DotIndexQ[aIndex]"
Dummy::usage           = "Dummy[<iClassQ>]"
DummyIndexQ::usage     = "DummyIndexQ[aIndex]"
DummyPair::usage       = "DummyPair[<iClassQ>]"
FlipIndex::usage       = "FlipIndex[aIndex, <testQ>]"
GetPairs::usage        = "GetPairs[indexL, <testQ>]"
GreekIndexQ::usage     = "GreekIndexQ[aIndex]"
HatIndexQ::usage       = "HatIndexQ[aIndex]"
IndexOrderedQ::usage   = "IndexOrderedQ[aIndex, bIndex] or IndexOrderedQ[{a,...}] or IndexOrderedQ[{a,...}, {b,...}]"
NormalIndexQ::usage    = "NormalIndexQ[aIndex]"
OneIndexQ::usage       = "OneIndexQ[aIndex]"
PairQ::usage           = "PairQ[aIndex, bIndex_, <testQ>] or PairQ[pairs, <testQ>]"
RegularIndexQ::usage   = "RegularIndexQ[aIndex]"
TensorialIndexQ::usage = "TensorialIndexQ[aIndex, <testQ>]"
ToDnIndex::usage       = "ToDnIndex[aIndex, <testQ>]"
ToUpIndex::usage       = "ToUpIndex[aIndex, <testQ>]"
UpIndexQ::usage        = "UpIndexQ[aIndex, <testQ>]"

(* SetIndexClass *)
MainIndexQ::usage    = "MainIndexQ[aIndex]"
SetIndexClass::usage = "SetIndexClass[iMainClassL, <iSubClassL>]"
SubIndexQ::usage     = "SubIndexQ[aIndex]"
ValidIndexQ::usage   = "ValidIndexQ[aIndex]"
ValidIndicesQ::usage = "ValidIndicesQ[aIndexL]"

(* DefineObject *)
ClearName::usage        = "ClearName[oName]"
DefinedNameQ::usage     = "DefinedNameQ[oName]"
DefinedTensorQ::usage   = "DefinedTensorQ[oName]"
DefineTensor::usage     = "DefineTensor[oName, permStr]"
PermWeight::usage       = "PermWeight[permW]"
SetAntisymmetric::usage = "SetAntisymmetric[oName]"
SetSymmetric::usage     = "SetSymmetric[oName]"
Symmetries::usage       = "Symmetries[oName]"

(* DefineOperator *)
DefinedOperatorQ::usage = "DefinedOperatorQ[opName]"
DefineOperator::usage   = "DefineOperator[opName, prtForm, opType]"
FormatOperatorXP::usage = ""

(* IndicalAtom *)
ExpandIAtom::usage     = "ExpandIAtom[expr]"
ExtractIAtom::usage    = "ExtractIAtom[expr]"
FreeIAtomQ::usage      = "FreeIAtomQ[expr]"
IAtomQ::usage          = "IAtomQ[expr]"
OrderingIndices::usage = "OrderingIndices[expr]"
WrapIAtom::usage       = "WrapIAtom[expr, wrapSymb]"

(* GetIndices *)
GetFreeIndices::usage = "GetFreeIndices[expr, <testQ>]"
GetIndices::usage     = "GetIndices[expr, <testQ>]"
GetMainIndices::usage = "GetMainIndices[expr, <testQ>]"
GetSubIndices::usage  = "GetSubIndices[expr, <testQ>]"

(* for Dummies *)
Dum::usage          = "Dum[expr]"
DumFresh::usage     = "DumFresh[expr]"
ResetDummies::usage = "ResetDummies[expr]"
SplitDum::usage     = "SplitDum[args, expr]"
SumDum::usage       = "SumDum[args, expr]"

(* Symmetrize *)
Antisymmetrize::usage = "Antiymmetrize[{a,..}, expr]"
Symmetrize::usage     = "Symmetrize[{a,..}, expr]"

(* ApplyRules *)
ApplyRules::usage = "ApplyRules[expr, rules]"
ObjectRule::usage = "ObjectRule[lhs, rhs, <cond>]"
ObjectSet::usage  = "ObjectSet[lhs, rhs, <cond>]"

(* SyntaxCheck *)
SyntaxCheck::usage = "SyntaxCheck[expr]"

(* KKTensor *)
SumDumKK::usage = "SumDumKK[expr]"
TableKK::usage  = "TableKK[expr, args]"

(* Dnup *)
Dnup::usage = "DnUp[expr, <covOpL>]"

(* Absorb *)
Absorb::usage = "Absorb[expr, aMetric, <covOpL>]"

(* IndexNotation *)
Can::usage = "Can[expr, <covOpL>]"

(***** Exported Variables *****)
(* Operator Type *)
CD::usage = ""
LD::usage = ""
TP::usage = ""
XD::usage = ""
XP::usage = ""

(* Flags *)
AutoCanFlag::usage     = ""
SyntaxCheckFlag::usage = ""

Begin["`Private`"]

(***************** Definition of Local Constants ********************)
defaultCDTable = {}  (* for Absorb and Can *)

iPredefinedNameTable = {C, D, E}  (* pre-defined in Mathematica *)
(********************************************************************)

(***************** Definition of Local Variables ********************)
(*** Flag Table *)
iFlagTable[AutoCanFlag]     = False
iFlagTable[SyntaxCheckFlag] = False
(********************************************************************)

(************************** Load Library ****************************)
<< dhPark`IndexNotationLib`

(***** for Indices *****)
Dummy[___] := (Message[Dummy::usage]; "")

DummyPair[___] := (Message[DummyPair::usage]; {"", ""})

forPrintIndex[args___] := {args, " "}  (* error *)

indexClassQ[___] := errorIndexQ

(***** for DefineTensor *****)
SetAntisymmetric[___] := Message[SetAntisymmetric::usage]
SetSymmetric[___]     := Message[SetSymmetric::usage]
Symmetries[___]       := Message[Symmetries::usage]
vanishingObjectQ[___] := False

(***** for IndicalAtom *****)
orderingIndicesObject[expr_] := expr

(***** for GetIndices *****)
getOriginalIndicesTermFor[expr_, testQ_] := {}

indicesSatisfyingFor[expr_, testQL_] := {}

(***** for Dummies *****)
SumDum[___] := Message[SumDum::usage]

(***** for SyntaxCheck *****)
syntaxCheckObject[expr_] := expr

(***** for DnUp *****)
canDnupQ[___] := False

flattenObject[___] := {}  (* non-indical *)

(***** for Absorb *****)
canPairQ[head_, _, _] := False

canAbsorbQ[expr_, covOpL_, {a_, b_}] := False

(***** for Can *****)
symToStrOneIndex[expr_] := expr /. {"l" <> oneIndices -> ToExpression["l" <> oneIndices],
                                    "u" <> oneIndices -> ToExpression["u" <> oneIndices]}
(********************************************************************)

(************************* for On / Off *****************************)
(* See Maeder, Programming in Mathematica, pp184) *)
postSave    = .      (* saved value of user's $Post *)
isAutoCanOn = False  (* whether AutoCan is currently turned on *)

(* Off tag *)
Unprotect[Off]  (* turn off a flag *)
Off[AutoCanFlag]     := If [ isAutoCanOn, $Post = postSave; isAutoCanOn = False; ];
Off[SyntaxCheckFlag] := Module[ {}, iFlagTable[SyntaxCheckFlag] = False; ]
Protect[Off]

(* On tag *)
Unprotect[On]  (* turn on a flag *)
On[AutoCanFlag]     :=
   If [!isAutoCanOn, postSave = If [ValueQ[$Post], $Post, Identity];
      isAutoCanOn = True;
      $Post = Can;
   ];
On[SyntaxCheckFlag] := Module[ {}, iFlagTable[SyntaxCheckFlag] = True; On[AutoCanFlag]]
Protect[On]
(********************************************************************)

End[ ] (* end the private context *)

EndPackage[ ]  (* end the package context *)

(*************** Unprotect Exported Symbols / Rutines ***************)
Protect[l0,la,lb,lc,ld,le,lf,lg,lh,li,lj,lk,ll,lm,ln,lo];
Protect[u0,ua,ub,uc,ud,ue,uf,ug,uh,ui,uj,uk,ul,um,un,uo];
Protect[lalpha,lbeta,lgamma,ldelta,lmu,lnu,lrho,lsigma,lepsilon,lkappa,llambda,ltau];
Protect[ualpha,ubeta,ugamma,udelta,umu,unu,urho,usigma,uepsilon,ukappa,ulambda,utau];
Protect[lA,lB,lC,lD,lE,lF,lG,lH,lI,lJ,lK,lL,lM,lN,lO];
Protect[uA,uB,uC,uD,uE,uF,uG,uH,uI,uJ,uK,uL,uM,uN,uO];
Protect[lha,lhb,lhc,lhd,lhe,lhf,lhg,lhh,lhi,lhj,lhk,lhl,lhm,lhn,lho];
Protect[uha,uhb,uhc,uhd,uhe,uhf,uhg,uhh,uhi,uhj,uhk,uhl,uhm,uhn,uho];
Protect[lba,lbb,lbc,lbd,lbe,lbf,lbg,lbh,lbi,lbj,lbk,lbl,lbm,lbn,lbo];
Protect[uba,ubb,ubc,ubd,ube,ubf,ubg,ubh,ubi,ubj,ubk,ubl,ubm,ubn,ubo];
Protect[lda,ldb,ldc,ldd,lde,ldf,ldg,ldh,ldi,ldj,ldk,ldl,ldm,ldn,ldo];
Protect[uda,udb,udc,udd,ude,udf,udg,udh,udi,udj,udk,udl,udm,udn,udo];
(********************************************************************)

Off[General::spell1]
Off[General::spell]
