(* :Title: KKTheory.m *)
(* :Context: dhPark`KKTheory` *)
(* :Author: D. H. Park *)
(* :Summary: for Kaluza-Klein Tensor *)
(* :Package Version: 1.0 *)
(* :Mathematica Version: 4.0 *)
(* :History:
   00.01.31: Initiate using IndexNotation
 *)

<< dhPark`IndexNotation`
<< dhPark`dhUtilities`

(********************** SetIndexClass ***********************)
SetIndexClass[CapitalIndexQ, {GreekIndexQ, NormalIndexQ}]  (* KKClass *)

(********************* Define Tensors ***********************)
DefineTensor[RiemannR, "R", "-bacd-abdc+cdab"]
DefineTensor[RicciR, "R", "ba"]
DefineTensor[ScalarR, "R", ""]
DefineTensor[AffineG, "\[CapitalGamma]", "abc"]  (* non-coordinate basis *)
DefineTensor[Metricg, "g", "ba"]
DefineTensor[Structuref, "f", "-bac"]
DefineTensor[GaugeF, "F", "-bac"]
DefineTensor[GaugeA, "A", "ab"]

DefineTensor[BaseRiemannR, "OverTilde[R]", "-bacd-abdc+cdab"]
DefineTensor[BaseRicciR, "OverTilde[R]", "ba"]
DefineTensor[BaseScalarR, "OverTilde[R]", ""]
DefineTensor[BaseAffineG, "OverTilde[\[CapitalGamma]]", "abc"]  (* coordinate basis *)
DefineTensor[BaseMetricg, "\[Gamma]", "ba"]
DefineTensor[BaseStructuref, "OverTilde[f]", "-bac"]

DefineTensor[FibreRiemannR, "OverHat[R]", "-bacd-abdc+cdab"]
DefineTensor[FibreRicciR, "OverHat[R]", "ba"]
DefineTensor[FibreScalarR, "OverHat[R]", ""]
DefineTensor[FibreAffineG, "OverHat[\[CapitalGamma]]", "abc"]  (* non-coordinate basis *)
DefineTensor[FibreMetricg, "\[Phi]", "ba"]
DefineTensor[FibreStructuref, "OverHat[f]", "-bac"]

(********************** Ordinary Derivative ************************)
DefineOperator[BD, "D", CD]

BD[opIndex_, moreIndices__, expr_] := Fold[BD[#2, #1]&, expr, Reverse[{opIndex, moreIndices}]]  \
                                      /; FreeQ[{opIndex, moreIndices, expr}, Pattern | Blank]
BD[opIndex_, aPlus_Plus]           := Map[BD[opIndex, #]&, aPlus]  \
                                      /; FreeQ[{opIndex, aPlus}, Pattern | Blank]
BD[opIndex_, aTimes_Times]         :=
    Module[ {expr = ExpandIAtom[aTimes], rcL},
        Switch [Head[expr],
            Plus,  Map[BD[opIndex, #]&, expr],
            Times, rcL = Map[BD[opIndex, #]&, List @@ expr];
                   Plus @@ Table[ReplacePart[expr, rcL[[i]], i], {i, 1, Length[expr]}],
            _,     BD[opIndex, expr]
        ]
    ] /; FreeQ[{opIndex, aTimes}, Pattern | Blank]
BD[opIndex_, E^x_]                 := E^x * BD[opIndex, x] /; FreeQ[{opIndex, x}, Pattern | Blank]
BD[opIndex_, x_^n_]                := n * x^(n-1) * BD[opIndex, x] /; FreeQ[{opIndex, x, n}, Pattern | Blank]
BD[opIndex_, n_?NumberQ]           := 0 /; FreeQ[opIndex, Pattern | Blank]
BD[opIndex_, c_Symbol]             := 0 /; FreeQ[opIndex, Pattern | Blank]  \
                                           && MemberQ[Attributes[c], Constant]

(******************* Block Diagonal Basis *********************)
SetAttributes[{g, \[Kappa]}, Constant]

BlockMetricRule = {
   ObjectRule[Metricg[la_, lb_], FibreMetricg[la, lb], AllQ[{la, lb}, NormalIndexQ]],
   ObjectRule[Metricg[la_, ub_], BaseMetricg[la, ub], AllQ[{la, ub}, GreekIndexQ]],
   ObjectRule[Metricg[la_, lb_], 0,
              (GreekIndexQ[la] && NormalIndexQ[lb]) || (GreekIndexQ[lb] && NormalIndexQ[la])]
}

BlockStructureRule = {
   ObjectRule[Structuref[la_, lb_, uc_], -g \[Kappa] GaugeF[la, lb, uc],
              GreekIndexQ[la] && GreekIndexQ[lb] && NormalIndexQ[uc]],
   ObjectRule[Structuref[la_, lb_, uc_], 0,
              ((GreekIndexQ[la] && NormalIndexQ[lb]) || (GreekIndexQ[lb] && NormalIndexQ[la])  \
               || AllQ[{la, lb}, GreekIndexQ] || AllQ[{la, lb}, NormalIndexQ])]
}

BlockDiagonalRule = Join[BlockMetricRule, BlockStructureRule];

(*********************** for AffineToMetric *************************)
(* tag AffineToMetric *)
AffineToMetric[AffineG[a_, b_, c_]] :=
   Module[ {rc, p, q, r, i},
      Switch [DnIndexQ /@ {a, b, c},
         {True, True, True},
            rc = dndndnAffineG[a, b, c],
         {True, True, False},
            p = Dummy[CapitalIndexQ];
            rc = Metricg[c, FlipIndex[p]] * dndndnAffineG[a, b, p],
         {True, False, True},
            p = Dummy[CapitalIndexQ];
            rc = Metricg[b, FlipIndex[p]] * dndndnAffineG[a, p, c],
         {True, False, False},
            {p, q} = {Dummy[CapitalIndexQ], Dummy[CapitalIndexQ]};
            rc = Metricg[b, FlipIndex[p]] * Metricg[c, FlipIndex[q]] * dndndnAffineG[a, p, q],
         {False, True, True},
            p = Dummy[CapitalIndexQ];
            rc = Metricg[a, FlipIndex[p]] * dndndnAffineG[p, b, c],
         {False, True, False},
            {p, q} = {Dummy[CapitalIndexQ], Dummy[CapitalIndexQ]};
            rc = Metricg[a, FlipIndex[p]] * Metricg[c, FlipIndex[q]] * dndndnAffineG[p, b, q],
         {False, False, True},
            {p, q} = {Dummy[CapitalIndexQ], Dummy[CapitalIndexQ]};
            rc = Metricg[a, FlipIndex[p]] * Metricg[b, FlipIndex[q]] * dndndnAffineG[p, q, c],
         {False, False, False},
            {p, q, r} = {Dummy[CapitalIndexQ], Dummy[CapitalIndexQ], Dummy[CapitalIndexQ]};
            rc = Metricg[a, FlipIndex[p]] * Metricg[b, FlipIndex[q]] * Metricg[c, FlipIndex[r]]
                 * dndndnAffineG[p, q, r]
      ];
      rc
   ]
AffineToMetric[BD[a_, expr_]] := BD[a, AffineToMetric[expr]]
AffineToMetric[expr_]         := expr

   (* where *)
   dndndnAffineG[a_, b_, c_] :=
      ( BD[a, Metricg[b, c]] + BD[b, Metricg[a, c]] - BD[c, Metricg[a, b]]
        + Structuref[a, b, c] + Structuref[c, a, b] + Structuref[c, b, a] ) / 2
(********************************************************************)

(********************* for RiemannToAffine **************************)
(* tag RiemannToAffine *)
RiemannToAffine[BD[a_,  expr_]]           := BD[a, RiemannToAffine[expr]]
RiemannToAffine[RiemannR[a_, b_, c_, d_]] := riemannToAffineAux[a, b, c, d]
RiemannToAffine[RicciR[a_, b_]]           :=
   Module[ {dummyDn, dummyUp},
      {dummyDn, dummyUp} = DummyPair[CapitalIndexQ];
      riemannToAffineAux[a, dummyDn, b, dummyUp]
   ]
RiemannToAffine[ScalarR[]]                :=
   Module[ {dummyDn1, dummyUp1, dummyDn2, dummyUp2},
      {dummyDn1, dummyUp1} = DummyPair[CapitalIndexQ];
      {dummyDn2, dummyUp2} = DummyPair[CapitalIndexQ];
      riemannToAffineAux[dummyDn1, dummyDn2, dummyUp1, dummyUp2]
   ]
RiemannToAffine[expr_] := expr

   (* where *)
   riemannToAffineAux[a_, b_, c_, d_] :=  (* ordering to {dn, up, dn, up} *)
      -riemannToAffineAux[b, a, c, d] /; UpIndexQ[a] && DnIndexQ[b]
   riemannToAffineAux[a_, b_, c_, d_] :=  (* ordering to {dn, up, dn, up} *)
      -riemannToAffineAux[a, b, d, c] /; UpIndexQ[c] && DnIndexQ[d]
   riemannToAffineAux[a_, b_, c_, d_] :=
      Module[ {rc, p, q, r},
         Switch [Map[DnIndexQ, {a, b, c, d}],
            {True, True, True, False},
               rc = dndndnupRiemannR[a, b, c, d],
            {True, True, True, True},
               p = Dummy[CapitalIndexQ];
               rc = Metricg[d, p] * dndndnupRiemannR[a, b, c, FlipIndex[p]],
            {True, True, False, False},
               p = Dummy[CapitalIndexQ];
               rc = Metricg[c, FlipIndex[p]] * dndndnupRiemannR[a, b, p, d],
            {True, False, True, False},
               p = Dummy[CapitalIndexQ];
               rc = Metricg[b, FlipIndex[p]] * dndndnupRiemannR[a, p, c, d],
            {True, False, True, True},
               rc = dndndnupRiemannR[c, d, a, b],
            {True, False, False, False},
               p = Dummy[CapitalIndexQ]; q = Dummy[CapitalIndexQ];
               rc = Metricg[b, FlipIndex[p]] * Metricg[c, FlipIndex[q]] * dndndnupRiemannR[a, p, q, d],
            {False, False, True, False},
               p = Dummy[CapitalIndexQ]; q = Dummy[CapitalIndexQ];
               rc = Metricg[a, FlipIndex[p]] * Metricg[b, FlipIndex[q]] * dndndnupRiemannR[p, q, c, d],
            {False, False, True, True},
               p = Dummy[CapitalIndexQ];
               rc = Metricg[a, FlipIndex[p]] * dndndnupRiemannR[c, d, p, b],
            {False, False, False, False},
               p = Dummy[CapitalIndexQ]; q = Dummy[CapitalIndexQ]; r = Dummy[CapitalIndexQ];
               rc = Metricg[a, FlipIndex[p]] * Metricg[b, FlipIndex[q]] * Metricg[c, FlipIndex[r]]
                    * dndndnupRiemannR[p, q, r, d]
         ];
         rc
      ]

      (* where *)
      dndndnupRiemannR[a_, b_, c_, d_] :=
         Module[ {dummyDn, dummyUp},
            {dummyDn, dummyUp} = DummyPair[CapitalIndexQ];

              BD[a, AffineG[b, c, d]] + AffineG[a, dummyDn, d] * AffineG[b, c, dummyUp]
            - BD[b, AffineG[a, c, d]] - AffineG[b, dummyDn, d] * AffineG[a, c, dummyUp]
            - Structuref[a, b, dummyUp] * AffineG[dummyDn, c, d]
         ]
(********************************************************************)
