(* :Title: Linealization.m *)
(* :Context: dhPark`Linealization` *)
(* :Author: D. H. Park *)
(* :Summary: for linealization in General Relativity *)
(* :Package Version: 1.0 *)
(* :Mathematica Version: 4.0 *)
(* :History:
   00.03.29: Initiate using IndexNotation
 *)
(* TODO
 *)

<< dhPark`IndexNotation`
<< dhPark`dhUtilities`

(********************* Define Tensorss ************************)
DefineTensor[RiemannR, "R", "-bacd-abdc+cdab"]
DefineTensor[RicciR, "R", "ba"]
DefineTensor[ScalarR, "R", ""]
DefineTensor[AffineG, "\[CapitalGamma]", "bac"]  (* coordinate basis *)
DefineTensor[Metricg, "g", "ba"]

(******************** Ordinary Derivative **********************)
DefineOperator[OD, "\[PartialD]", CD]

OD[opIndex_, moreIndices__, expr_] := Fold[OD[#2, #1]&, expr, Reverse[{opIndex, moreIndices}]]  \
                                      /; FreeQ[{opIndex, moreIndices, expr}, Pattern | Blank]
OD[opIndex_, aPlus_Plus]           := Map[OD[opIndex, #]&, aPlus]  \
                                      /; FreeQ[{opIndex, aPlus}, Pattern | Blank]
OD[opIndex_, aTimes_Times]         :=
    Module[ {expr = ExpandIAtom[aTimes], rcL},
        Switch [Head[expr],
            Plus,  Map[OD[opIndex, #]&, expr],
            Times, rcL = Map[OD[opIndex, #]&, List @@ expr];
                   Plus @@ Table[ReplacePart[expr, rcL[[i]], i], {i, 1, Length[expr]}],
            _,     OD[opIndex, expr]
        ]
    ] /; FreeQ[{opIndex, aTimes}, Pattern | Blank]
OD[a_, OD[b_, expr_]]              := OD[b, OD[a, expr]]  \
                                      /; FreeQ[{a, b, expr}, Pattern | Blank]  \
                                         && DnIndexQ[a] && DnIndexQ[b] && !IndexOrderedQ[a,b]
OD[opIndex_, E^x_]                 := E^x * OD[opIndex, x]  \
                                      /; FreeQ[{opIndex, x}, Pattern | Blank]
OD[opIndex_, x_^n_]                := n * x^(n-1) * OD[opIndex, x]  \
                                      /; FreeQ[{opIndex, x, n}, Pattern | Blank]
OD[opIndex_, n_?NumberQ]           := 0 /; FreeQ[opIndex, Pattern | Blank]
OD[opIndex_, c_Symbol]             := 0 /; FreeQ[opIndex, Pattern | Blank]  \
                                           && MemberQ[Attributes[c], Constant]

(*********************** for AffineToMetric *************************)
AffineToMetric[AffineG[a_, b_, c_]] :=
   Module[ {rc, p, q, r, i},
      Switch [DnIndexQ /@ {a, b, c},
         {True, True, True},
            rc = dndndnAffineG[a, b, c],
         {True, True, False},
            p = Dummy[NormalIndexQ];
            rc = Metricg[c, FlipIndex[p]] * dndndnAffineG[a, b, p],
         {True, False, True},
            p = Dummy[NormalIndexQ];
            rc = Metricg[b, FlipIndex[p]] * dndndnAffineG[a, p, c],
         {True, False, False},
            {p, q} = {Dummy[NormalIndexQ], Dummy[NormalIndexQ]};
            rc = Metricg[b, FlipIndex[p]] * Metricg[c, FlipIndex[q]] * dndndnAffineG[a, p, q],
         {False, True, True},
            p = Dummy[NormalIndexQ];
            rc = Metricg[a, FlipIndex[p]] * dndndnAffineG[p, b, c],
         {False, True, False},
            {p, q} = {Dummy[NormalIndexQ], Dummy[NormalIndexQ]};
            rc = Metricg[a, FlipIndex[p]] * Metricg[c, FlipIndex[q]] * dndndnAffineG[p, b, q],
         {False, False, True},
            {p, q} = {Dummy[NormalIndexQ], Dummy[NormalIndexQ]};
            rc = Metricg[a, FlipIndex[p]] * Metricg[b, FlipIndex[q]] * dndndnAffineG[p, q, c],
         {False, False, False},
            {p, q, r} = {Dummy[NormalIndexQ], Dummy[NormalIndexQ], Dummy[NormalIndexQ]};
            rc = Metricg[a, FlipIndex[p]] * Metricg[b, FlipIndex[q]] * Metricg[c, FlipIndex[r]]
                 * dndndnAffineG[p, q, r]
      ];
      rc
   ]
AffineToMetric[OD[a_, expr_]] := OD[a, AffineToMetric[expr]]
AffineToMetric[expr_]         := expr

   (* where *)
   dndndnAffineG[a_, b_, c_] :=
      ( OD[a, Metricg[b, c]] + OD[b, Metricg[a, c]] - OD[c, Metricg[a, b]] ) / 2
(********************************************************************)

(********************* for RiemannToAffine **************************)
RiemannToAffine[OD[a_,  expr_]]           := OD[a, RiemannToAffine[expr]]
RiemannToAffine[RiemannR[a_, b_, c_, d_]] := riemannToAffineAux[a, b, c, d]
RiemannToAffine[RicciR[a_, b_]]           :=
   Module[ {dummyDn, dummyUp},
      {dummyDn, dummyUp} = DummyPair[NormalIndexQ];
      riemannToAffineAux[a, dummyDn, b, dummyUp]
   ]
RiemannToAffine[ScalarR[]]                :=
   Module[ {dummyDn1, dummyUp1, dummyDn2, dummyUp2},
      {dummyDn1, dummyUp1} = DummyPair[NormalIndexQ];
      {dummyDn2, dummyUp2} = DummyPair[NormalIndexQ];
      riemannToAffineAux[dummyDn1, dummyDn2, dummyUp1, dummyUp2]
   ]
RiemannToAffine[expr_] := expr

   (* where *)
   riemannToAffineAux[a_, b_, c_, d_] :=  (* ordering to {dn, up, dn, up} *)
      -riemannToAffineAux[b, a, c, d] /; UpIndexQ[a] && DnIndexQ[b]
   riemannToAffineAux[a_, b_, c_, d_] :=  (* ordering to {dn, up, dn, up} *)
      -riemannToAffineAux[a, b, d, c] /; UpIndexQ[c] && DnIndexQ[d]
   riemannToAffineAux[a_, b_, c_, d_] :=
      Module[ {rc, p, q, r},
         Switch [Map[DnIndexQ, {a, b, c, d}],
            {True, True, True, False},
               rc = dndndnupRiemannR[a, b, c, d],
            {True, True, True, True},
               p = Dummy[NormalIndexQ];
               rc = Metricg[d, p] * dndndnupRiemannR[a, b, c, FlipIndex[p]],
            {True, True, False, False},
               p = Dummy[NormalIndexQ];
               rc = Metricg[c, FlipIndex[p]] * dndndnupRiemannR[a, b, p, d],
            {True, False, True, False},
               p = Dummy[NormalIndexQ];
               rc = Metricg[b, FlipIndex[p]] * dndndnupRiemannR[a, p, c, d],
            {True, False, True, True},
               rc = dndndnupRiemannR[c, d, a, b],
            {True, False, False, False},
               p = Dummy[NormalIndexQ]; q = Dummy[NormalIndexQ];
               rc = Metricg[b, FlipIndex[p]] * Metricg[c, FlipIndex[q]] * dndndnupRiemannR[a, p, q, d],
            {False, False, True, False},
               p = Dummy[NormalIndexQ]; q = Dummy[NormalIndexQ];
               rc = Metricg[a, FlipIndex[p]] * Metricg[b, FlipIndex[q]] * dndndnupRiemannR[p, q, c, d],
            {False, False, True, True},
               p = Dummy[NormalIndexQ];
               rc = Metricg[a, FlipIndex[p]] * dndndnupRiemannR[c, d, p, b],
            {False, False, False, False},
               p = Dummy[NormalIndexQ]; q = Dummy[NormalIndexQ]; r = Dummy[NormalIndexQ];
               rc = Metricg[a, FlipIndex[p]] * Metricg[b, FlipIndex[q]] * Metricg[c, FlipIndex[r]]
                    * dndndnupRiemannR[p, q, r, d]
         ];
         rc
      ]

      (* where *)
      dndndnupRiemannR[a_, b_, c_, d_] :=
         Module[ {dummyDn, dummyUp},
            {dummyDn, dummyUp} = DummyPair[NormalIndexQ];

              OD[a, AffineG[b, c, d]] + AffineG[a, dummyDn, d] * AffineG[b, c, dummyUp]
            - OD[b, AffineG[a, c, d]] - AffineG[b, dummyDn, d] * AffineG[a, c, dummyUp]
         ]
(********************************************************************)
